//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
//%      Agumented tree method for image matching (CVPR2011)   %
//%             Hao Jiang at Boston College                    %
//%                   Oct. 2010                                %
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/////////////////////////////////////////////////////////////////
//  Decompose for cos - sin = 1,  0 =< cos <= 1,  -1 =< sin <= 0 
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <glpk.h>
#include <math.h>

LPX *lpc;
double *x;

const int TARGET_NUM = 3;
const int SRC_NUM = 3;

const int VN = TARGET_NUM*SRC_NUM+2;  // Useful variables number
const int MAX_ITERS = 2000;
const int PAIR_NUM = SRC_NUM-1;
const int CST = 3*PAIR_NUM; // # of complex constraints

double sol[VN+1][MAX_ITERS+1];

double edgex[PAIR_NUM+1][TARGET_NUM+1][TARGET_NUM+1]; // Org objective coefs
double pcoefx[PAIR_NUM+1];
double qcoefx[PAIR_NUM+1];
double hcoefx[PAIR_NUM+1];
double gcoefx[PAIR_NUM+1];
double ucoefx[PAIR_NUM+1];
double vcoefx[PAIR_NUM+1];
double cos_coefx;
double sss_coefx;

double edge[PAIR_NUM+1][TARGET_NUM+1][TARGET_NUM+1]; // Simple LP objective coefs
double pcoef[PAIR_NUM+1];
double qcoef[PAIR_NUM+1];
double hcoef[PAIR_NUM+1];
double gcoef[PAIR_NUM+1];
double ucoef[PAIR_NUM+1];
double vcoef[PAIR_NUM+1];
double cos_coef;
double sss_coef;

double cx_eg[CST+1][TARGET_NUM+1][TARGET_NUM+1]; // Complex constraint coefs


int pair[PAIR_NUM+1][3];
double cost[SRC_NUM+1][TARGET_NUM+1];
double w[MAX_ITERS+1]; // the weights of proposals

void init_lpc()
{
      lpc = lpx_create_prob();
      lpx_set_obj_dir(lpc, LPX_MAX);
      for (int i = 1; i <= PAIR_NUM; i++) // cos constraints  
      {
          lpx_add_rows(lpc, 1);
          lpx_set_row_bnds(lpc, i, LPX_FX, 0.0, 0.0);
      }

      for (int i = 1; i <= PAIR_NUM; i++) // sin constraints  
      {
          lpx_add_rows(lpc, 1);
          lpx_set_row_bnds(lpc, i+PAIR_NUM, LPX_FX, -1.0, -1.0);
      }

      for (int i = 1; i <= PAIR_NUM; i++) // scale constraints  
      {
          lpx_add_rows(lpc, 1);
          lpx_set_row_bnds(lpc, i+2*PAIR_NUM, LPX_FX, 0.0, 0.0);
      }


      lpx_add_rows(lpc, 1);
      lpx_set_row_bnds(lpc, CST+1, LPX_FX, 1.0, 1.0);
}

void update_lpc(double obj, double cxv[])
{
      static int num_col = 1;

      lpx_add_cols(lpc, 1);
      lpx_set_col_bnds(lpc, num_col, LPX_LO, 0.0, 0.0);
      lpx_set_obj_coef(lpc, num_col, obj);

      int ind[CST+2];
      double val[CST+2];

      int cn = 0;
      for (int i = 1; i <= CST; i++)
      {
         if (cxv[i] != 0)
         {
                 cn++;
                 ind[cn] = i;
                 val[cn] = cxv[i];
         }
      }

      ind[cn+1] = CST+1;
      val[cn+1] = 1.0;
      lpx_set_mat_col(lpc, num_col, cn+1, ind, val);

      num_col ++;
}

void update_lp_simp(double* u)
{
         for (int n = 1; n <= PAIR_NUM; n++) 
         for (int i = 1; i <= TARGET_NUM; i++)
         for (int j = 1; j <= TARGET_NUM; j++)
         {
             double s = 0;
             for (int m = 0; m < 3; m++)
                s = s + u[m*PAIR_NUM+n] * cx_eg[m*PAIR_NUM+n][i][j];

             edge[n][i][j] = edgex[n][i][j] - s;
         }

         for (int n = 1; n <= PAIR_NUM; n++)
         {
             pcoef[n] = pcoefx[n] - u[n] * (-1);
             qcoef[n] = qcoefx[n] - u[n] * (+1);
         }

         for (int n = 1; n <= PAIR_NUM; n++)
         {
             hcoef[n] = hcoefx[n] - u[n+PAIR_NUM] * (-1);
             gcoef[n] = gcoefx[n] - u[n+PAIR_NUM] * (+1);
         }

         for (int n = 1; n <= PAIR_NUM; n++)
         {
             ucoef[n] = ucoefx[n] - u[n+2*PAIR_NUM] * (-1);
             vcoef[n] = vcoefx[n] - u[n+2*PAIR_NUM] * (+1);
         }

         cos_coef = cos_coefx;
         sss_coef = sss_coefx;
         for (int n = 1; n <= PAIR_NUM; n++)
         {
             cos_coef = cos_coef - u[n] * (-1) -  u[n+PAIR_NUM] * (-1);
             sss_coef = sss_coef - u[n+2*PAIR_NUM] * (-1);
         }         
        
}
       

void ins_sol(double x[], int &num)
{
     for (int n = 1; n <= VN; n++)
        sol[n][num] = x[n];
     
     num ++;
}

void init_simple_lp()
{
      FILE *fpcost = fopen("cost.txt", "rt");
      for (int i = 1; i <= SRC_NUM; i++)
         for (int j = 1; j <= TARGET_NUM; j++)
         {
            float t;
            fscanf(fpcost, "%f", &t);
            cost[i][j] = t;
         } 
      fclose(fpcost);    

      FILE *fppair = fopen("pair.txt", "rt");
      for (int i = 1; i <= PAIR_NUM; i++)
         for (int j = 1; j <= 2; j++)
         {
            int t ;
            fscanf(fppair, "%d", &t);
            pair[i][j] = t;
         }
      fclose(fppair);
      
      for (int n = 1; n <= PAIR_NUM; n++)
      {
         int r = pair[n][1];
         int c = pair[n][2];
         for (int i = 1; i <= TARGET_NUM; i++)
             for (int j = 1; j <= TARGET_NUM; j++)
             {
               if (n == 1)
                  edge[n][i][j] = -(cost[r][i] + cost[c][j]);
               else
                  edge[n][i][j] = -cost[c][j];
             }
      }

      for (int n = 1; n <= PAIR_NUM; n++)
      {
         for (int i = 1; i <= TARGET_NUM; i++)
         {
             for (int j = 1; j <= TARGET_NUM; j++)
               edgex[n][i][j] = edge[n][i][j];
         }
      }


      for (int i = 1; i <= PAIR_NUM; i++)
      {
          pcoef[i] = -100;
          qcoef[i] = -100;
          hcoef[i] = -100;
          gcoef[i] = -100;
          ucoef[i] = -100;
          vcoef[i] = -100;

          pcoefx[i] = -100;
          qcoefx[i] = -100;
          hcoefx[i] = -100;
          gcoefx[i] = -100;
          ucoefx[i] = -100;
          vcoefx[i] = -100;
      }  

      cos_coef = 0;
      sss_coef = 0;

      cos_coefx = 0;
      sss_coefx = 0;

      ///////// Populate the complex constraint coefficients
      for (int n = 1; n <= CST; n++)
      {
         char s[10];
         sprintf(s, "cx%d.txt", n);
         FILE *fp = fopen(s, "rt");
         for (int i = 1; i <= TARGET_NUM; i++)
           for (int j = 1; j <= TARGET_NUM; j++)
           {
              float t;
              fscanf(fp, "%f", &t);
              cx_eg[n][i][j] = t;
           }
         fclose(fp);
      }

}

void sol_simple_lp(int maxid[], double pp[], double qq[], double hh[], double gg[], 
                   double uu[], double vv[], double &cosv, double &sssv, double &obj)  
{
      double maxs = -1e10;

      for (int i = 1; i <= TARGET_NUM; i++)
      {   
          double s = 0;
          int id[PAIR_NUM+1][TARGET_NUM+1];
                 
          for (int n = 1; n <= PAIR_NUM; n++)
          {     
               double max = -1e10;
               for (int j = 1; j <= TARGET_NUM; j++)
               {
                 if (edge[n][i][j] > max) 
                 {
                    max = edge[n][i][j];
                    id[n][i] = j;
                 }
               }
               s = s + max;
          }
          
          if (s > maxs)
          {
              maxs = s;
              for (int n = 1; n <= PAIR_NUM; n++)
              {
                   maxid[pair[n][1]] = i;
                   maxid[pair[n][2]] = id[n][i];
              }
          }
             
      }
 
      for (int i = 1; i <= PAIR_NUM; i++)
      {
         if ( pcoef[i] <= 0) pp[i] = 0;
          else pp[i] = 3;

         if ( qcoef[i] <= 0) qq[i] = 0;
          else qq[i] = 3;

         if ( hcoef[i] <= 0) hh[i] = 0;
          else hh[i] = 3;

         if ( gcoef[i] <= 0) gg[i] = 0;
          else gg[i] = 3;

         if ( ucoef[i] <= 0) uu[i] = 0;
          else uu[i] = 3;

         if ( vcoef[i] <= 0) vv[i] = 0;
          else vv[i] = 3;
      }

      if (cos_coef <= 0)
          cosv = 0;
       else
          cosv = 1;

      if (sss_coef <= 0)
          sssv = 0.5;
       else
          sssv = 2;


    double cc = 0;

    for (int i = 1; i <= PAIR_NUM; i++)
    {
         cc = cc + pp[i] * pcoef[i];
         cc = cc + qq[i] * qcoef[i];
         cc = cc + hh[i] * hcoef[i];
         cc = cc + gg[i] * gcoef[i];
         cc = cc + uu[i] * ucoef[i];
         cc = cc + vv[i] * vcoef[i];
    }

    cc = cc + cosv * cos_coef;
    cc = cc + sssv * sss_coef;

    obj = cc + maxs;
}

double get_obj(int tid[], double p[], double q[], double h[], double g[],
               double u[], double v[], double cosv, double sssv)
{
    double cc = 0;
    for (int i = 1; i <= SRC_NUM; i++)
         cc = cc - cost[i][tid[i]];

    for (int i = 1; i <= PAIR_NUM; i++)
    {
         cc = cc + p[i] * pcoefx[i];
         cc = cc + q[i] * qcoefx[i];
         cc = cc + h[i] * hcoefx[i];
         cc = cc + g[i] * gcoefx[i];
         cc = cc + u[i] * ucoefx[i];
         cc = cc + v[i] * vcoefx[i];
    }

    cc = cc + cosv * cos_coefx;
    cc = cc + sssv * sss_coefx;
    return cc;
}


double get_cx_constraint_values(int tid[], double p[], double q[], double h[], double g[],
                                double u[], double v[], double cosv, double sssv, double cxv[])
{
    double cc = 0;
    int cn = 1;
    for (int i = 1; i <= PAIR_NUM; i++)
    {
       int e1 = pair[i][1];
       int e2 = pair[i][2];
       double cc = cx_eg[i][tid[e1]][tid[e2]] - cosv - p[i] + q[i];
       cxv[cn] = cc;
       cn++;
    }
        
    for (int i = 1; i <= PAIR_NUM; i++)
    {
       int e1 = pair[i][1];
       int e2 = pair[i][2];
       double cc = cx_eg[i+PAIR_NUM][tid[e1]][tid[e2]] - cosv - h[i] + g[i];
       cxv[cn] = cc;
       cn++;       
    }

    for (int i = 1; i <= PAIR_NUM; i++)
    {
       int e1 = pair[i][1];
       int e2 = pair[i][2];
       double cc = cx_eg[i+2*PAIR_NUM][tid[e1]][tid[e2]] - sssv - u[i] + v[i];
       cxv[cn] = cc;
       cn++;
    }

}     
          
void groupsol(double x[], int tid[], double pp[], double qq[], double hh[], double gg[], double uu[], double vv[], 
              double cosv, double sssv)
{
   int cn = 1;
   for (int i = 1; i <= SRC_NUM; i++)
   for (int j = 1; j <= TARGET_NUM; j++)
   {
        x[cn] = (tid[i] == j);
        cn++;
   }


   x[cn] = cosv; cn++;
   x[cn] = sssv;
}
   



int main(void)
{    
      init_lpc();
      x = new double[VN+1];

      /////////////////// Prepare ///////////////////////
      init_simple_lp();
      //////////////////////////////////////////////////                  
      int tid[SRC_NUM+1];
      double pp[PAIR_NUM+1], qq[PAIR_NUM+1], hh[PAIR_NUM+1], gg[PAIR_NUM+1];
      double uu[PAIR_NUM+1], vv[PAIR_NUM+1];
      double cosv, sinv, sssv;
      double obj;
      sol_simple_lp(tid, pp, qq, hh, gg, uu, vv, cosv,  sssv, obj);

      for (int i = 1; i <= SRC_NUM; i++)
          printf("%d %d\n", i, tid[i]);

      for (int i = 1; i <= PAIR_NUM; i++)
      {
          printf("%f %f\n", pp[i], qq[i]);
          printf("%f %f\n", hh[i], gg[i]);
          printf("%f %f\n", uu[i], vv[i]);
      }

      printf("%f %f %f\n", cosv, sinv, sssv);

      printf("obj = %f\n", obj); 

      ////////////////////////////////////////////////////////////////////////////////////
      double cxv[CST+1];
      get_cx_constraint_values(tid, pp, qq, hh, gg, uu, vv, cosv, sssv, cxv);

      for (int i = 1; i <= CST; i++)
         printf("%f ", cxv[i]);
      
      double maxr = -1e10;
      for (int i = 1; i <= CST; i++)
      {
          if (fabs(cxv[i]) > maxr)
              maxr = fabs(cxv[i]);
      }
      maxr = maxr + 1;      
      //////////////////////////////////////////////////////////////////////////////
      int snum = 1;
      obj = get_obj(tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
      update_lpc(obj, cxv);
      groupsol(x, tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
      ins_sol(x, snum);
     
      for (int i = 1; i <= PAIR_NUM; i++)
      {
        pp[i] = maxr + cxv[i];
        qq[i] = 0;
      }

      for (int i = 1; i <= PAIR_NUM; i++)
      {
        hh[i] = maxr + cxv[i+PAIR_NUM] - (-1);
        gg[i] = 0;
      }

      for (int i = 1; i <= PAIR_NUM; i++)
      {
        uu[i] = maxr + cxv[i+2*PAIR_NUM];
        vv[i] = 0;
      }

      double cxtmp[CST+1];
      get_cx_constraint_values(tid, pp, qq, hh, gg, uu, vv, cosv, sssv, cxtmp);
      obj = get_obj(tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
      update_lpc(obj, cxtmp);
      groupsol(x, tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
      ins_sol(x, snum);
      ////////////////////////////////////////////////////////////////////////////////
      for (int i = 1; i <= PAIR_NUM; i++)
      {
        pp[i] = 0;
        qq[i] = maxr - cxv[i];
      }

      for (int i = 1; i <= PAIR_NUM; i++)
      {
        hh[i] = 0;
        gg[i] = maxr - cxv[i+PAIR_NUM] + (-1);
      }

      for (int i = 1; i <= PAIR_NUM; i++)
      {
        uu[i] = 0;
        vv[i] = maxr - cxv[i+2*PAIR_NUM];
      }

      get_cx_constraint_values(tid, pp, qq, hh, gg, uu, vv, cosv, sssv, cxtmp);
      obj = get_obj(tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
      update_lpc(obj, cxtmp);
      groupsol(x, tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
      ins_sol(x, snum);
     
      double z; 
      for (int n = 1; n <= MAX_ITERS; n++)
      {
        lpx_simplex(lpc);
        z = lpx_get_obj_val(lpc);
        for (int i = 1; i <= snum-1; i++)
        {
           w[i] = lpx_get_col_prim(lpc, i);
        }
  
        printf("opt value = %f\n", z);
        double u[CST+1];
        for (int i = 1; i <= CST; i++)
        {
           u[i] = lpx_get_row_dual(lpc, i);
        }
        
        double delta = lpx_get_row_dual(lpc, CST+1);
        update_lp_simp(u);

        sol_simple_lp(tid, pp, qq, hh, gg, uu, vv, cosv, sssv, obj);
        printf("\n[%d, %f %f]\n", n, obj, delta);

        if (fabs(obj - delta) < 0.00001) break;
        get_cx_constraint_values(tid, pp, qq, hh, gg, uu, vv, cosv, sssv, cxtmp);
        obj = get_obj(tid, pp, qq, hh, gg, uu, vv, cosv, sssv);  
        update_lpc(obj, cxtmp);
        groupsol(x, tid, pp, qq, hh, gg, uu, vv, cosv, sssv);
        ins_sol(x, snum);
      }

      FILE *fpsol = fopen("sol3.txt", "wt");
      int cn = 1;
      for (int i = 1; i <= SRC_NUM; i++)
       for ( int j = 1; j <= TARGET_NUM; j++)
       {
           double t = 0;
           for (int m = 1; m <= snum-1; m++)
              t +=  w[m] * sol[cn][m];
           cn++;
           fprintf(fpsol, "%d %d %f\n", i, j, t);
       }
      fclose(fpsol);

      double cosval = 0;
      for (int i = VN-1; i <= VN-1; i++)
      {
           for (int j = 1; j <= snum-1; j++)
              cosval +=  w[j] * sol[i][j];
      }
         
      FILE *fpopt = fopen("optima.txt", "wt");
      fprintf(fpopt, "%f", z);
      fclose(fpopt);

      FILE *fpcs = fopen("cossin.txt", "wt");
      fprintf(fpcs, "%f %f", cosval, cosval-1);
      fclose(fpcs);

      lpx_delete_prob(lpc);
      delete [] x;
      return 0;
}

